@@ -22,10 +22,9 @@ import java.io.*;
 import java.lang.reflect.*;
 import java.net.*;
 import java.security.*;
+import java.util.HashMap;
 import java.util.Hashtable;
-import java.util.jar.JarFile;
 
-import org.apache.felix.framework.searchpolicy.ModuleImpl;
 import org.osgi.framework.BundleActivator;
 import org.osgi.framework.BundleContext;
 
@@ -57,7 +56,23 @@ public class SecureAction
 
     public SecureAction()
     {
-        m_acc = AccessController.getContext();
+        if (System.getSecurityManager() != null)
+        {
+            try
+            {
+                Actions actions = (Actions) m_actions.get();
+                actions.set(Actions.INITIALIZE_CONTEXT, null);
+                m_acc = (AccessControlContext) AccessController.doPrivileged(actions);
+            }
+            catch (PrivilegedActionException ex)
+            {
+                throw (RuntimeException) ex.getException();
+            }
+        }
+        else
+        {
+            m_acc = AccessController.getContext();
+        }
     }
 
     public String getSystemProperty(String name, String def)
@@ -81,6 +96,69 @@ public class SecureAction
         }
     }
 
+    public ClassLoader getParentClassLoader(ClassLoader loader)
+    {
+        if (System.getSecurityManager() != null)
+        {
+            try
+            {
+                Actions actions = (Actions) m_actions.get();
+                actions.set(Actions.GET_PARENT_CLASS_LOADER_ACTION, loader);
+                return (ClassLoader) AccessController.doPrivileged(actions, m_acc);
+            }
+            catch (PrivilegedActionException ex)
+            {
+                throw (RuntimeException) ex.getException();
+            }
+        }
+        else
+        {
+            return loader.getParent();
+        }
+    }
+
+    public ClassLoader getSystemClassLoader()
+    {
+        if (System.getSecurityManager() != null)
+        {
+            try
+            {
+                Actions actions = (Actions) m_actions.get();
+                actions.set(Actions.GET_SYSTEM_CLASS_LOADER_ACTION);
+                return (ClassLoader) AccessController.doPrivileged(actions, m_acc);
+            }
+            catch (PrivilegedActionException ex)
+            {
+                throw (RuntimeException) ex.getException();
+            }
+        }
+        else
+        {
+            return ClassLoader.getSystemClassLoader();
+        }
+    }
+    
+    public ClassLoader getClassLoader(Class clazz)
+    {
+        if (System.getSecurityManager() != null)
+        {
+            try
+            {
+                Actions actions = (Actions) m_actions.get();
+                actions.set(Actions.GET_CLASS_LOADER_ACTION, clazz);
+                return (ClassLoader) AccessController.doPrivileged(actions, m_acc);
+            }
+            catch (PrivilegedActionException ex)
+            {
+                throw (RuntimeException) ex.getException();
+            }
+        }
+        else
+        {
+            return clazz.getClassLoader();
+        }
+    }
+
     public Class forName(String name) throws ClassNotFoundException
     {
         if (System.getSecurityManager() != null)
@@ -161,6 +239,27 @@ public class SecureAction
         }
     }
 
+    public Process exec(String command) throws IOException
+    {
+        if (System.getSecurityManager() != null)
+        {
+            try
+            {
+                Actions actions = (Actions) m_actions.get();
+                actions.set(Actions.EXEC_ACTION, command);
+                return (Process) AccessController.doPrivileged(actions, m_acc);
+            }
+            catch (PrivilegedActionException ex)
+            {
+                throw (RuntimeException) ex.getException();
+            }
+        }
+        else
+        {
+            return Runtime.getRuntime().exec(command);
+        }
+    }
+
     public String getAbsolutePath(File file)
     {
         if (System.getSecurityManager() != null)
@@ -463,33 +562,6 @@ public class SecureAction
         }
     }
 
-    public JarFile getJarURLConnectionJAR(JarURLConnection connection)
-        throws IOException
-    {
-        if (System.getSecurityManager() != null)
-        {
-            try
-            {
-                Actions actions = (Actions) m_actions.get();
-                actions.set(Actions.OPEN_JARURLCONNECTIONJAR_ACTION, connection);
-                return (JarFile) AccessController.doPrivileged(actions,
-                    m_acc);
-            }
-            catch (PrivilegedActionException ex)
-            {
-                if (ex.getException() instanceof IOException)
-                {
-                    throw (IOException) ex.getException();
-                }
-                throw (RuntimeException) ex.getException();
-            }
-        }
-        else
-        {
-            return connection.getJarFile();
-        }
-    }
-
     public JarFileX openJAR(File file) throws IOException
     {
         if (System.getSecurityManager() != null)
@@ -539,35 +611,7 @@ public class SecureAction
             return new JarFileX(file, verify);
         }
     }
-// TODO: REFACTOR - SecureAction fix needed.
-/*
-    public ModuleClassLoader createModuleClassLoader(ModuleImpl impl)
-    {
-        return createModuleClassLoader(impl, null);
-    }
 
-    public ModuleClassLoader createModuleClassLoader(ModuleImpl impl,
-        ProtectionDomain protectionDomain)
-    {
-        if (System.getSecurityManager() != null)
-        {
-            try
-            {
-                Actions actions = (Actions) m_actions.get();
-                actions.set(Actions.CREATE_MODULECLASSLOADER_ACTION, impl, protectionDomain);
-                return (ModuleClassLoader) AccessController.doPrivileged(actions, m_acc);
-            }
-            catch (PrivilegedActionException ex)
-            {
-                throw (RuntimeException) ex.getException();
-            }
-        }
-        else
-        {
-            return new ModuleClassLoader(impl, protectionDomain);
-        }
-    }
-*/
     public void startActivator(BundleActivator activator, BundleContext context)
         throws Exception
     {
@@ -679,6 +723,27 @@ public class SecureAction
         }
     }
 
+    public Constructor getDeclaredConstructor(Class target, Class[] types) throws Exception
+    {
+        if (System.getSecurityManager() != null)
+        {
+            Actions actions = (Actions) m_actions.get();
+            actions.set(Actions.GET_DECLARED_CONSTRUCTOR_ACTION, target, types);
+            try
+            {
+                return (Constructor) AccessController.doPrivileged(actions, m_acc);
+            }
+            catch (PrivilegedActionException e)
+            {
+                throw e.getException();
+            }
+        }
+        else
+        {
+            return target.getDeclaredConstructor(types);
+        }
+    }
+
     public Method getMethod(Class target, String method, Class[] types) throws Exception
     {
         if (System.getSecurityManager() != null)
@@ -705,7 +770,7 @@ public class SecureAction
         if (System.getSecurityManager() != null)
         {
             Actions actions = (Actions) m_actions.get();
-            actions.set(Actions.GET_DECLAREDMETHOD_ACTION, target, method, types);
+            actions.set(Actions.GET_DECLARED_METHOD_ACTION, target, method, types);
             try
             {
                 return (Method) AccessController.doPrivileged(actions, m_acc);
@@ -721,12 +786,12 @@ public class SecureAction
         }
     }
 
-    public void setAccesssible(Method method)
+    public void setAccesssible(AccessibleObject ao)
     {
         if (System.getSecurityManager() != null)
         {
             Actions actions = (Actions) m_actions.get();
-            actions.set(Actions.SET_ACCESSIBLE_ACTION, method);
+            actions.set(Actions.SET_ACCESSIBLE_ACTION, ao);
             try
             {
                 AccessController.doPrivileged(actions, m_acc);
@@ -738,7 +803,7 @@ public class SecureAction
         }
         else
         {
-            method.setAccessible(true);
+            ao.setAccessible(true);
         }
     }
 
@@ -763,7 +828,7 @@ public class SecureAction
             return method.invoke(target, params);
         }
     }
-    
+
     public Object invokeDirect(Method method, Object target, Object[] params) throws Exception
     {
         if (System.getSecurityManager() != null)
@@ -925,43 +990,102 @@ public class SecureAction
         return null;
     }
 
+    public void flush(Class targetClazz, Object lock) throws Exception
+    {
+        if (System.getSecurityManager() != null)
+        {
+            Actions actions = (Actions) m_actions.get();
+            actions.set(Actions.FLUSH_FIELD_ACTION, targetClazz, lock);
+            try
+            {
+                AccessController.doPrivileged(actions, m_acc);
+            }
+            catch (PrivilegedActionException e)
+            {
+                throw e.getException();
+            }
+        }
+        else
+        {
+            _flush(targetClazz, lock);
+        }
+    }
+
+    private static void _flush(Class targetClazz, Object lock) throws Exception
+    {
+        synchronized (lock) 
+        {
+            Field[] fields = targetClazz.getDeclaredFields();
+            // reset cache
+            for (int i = 0; i < fields.length; i++)
+            {
+                if (Modifier.isStatic(fields[i].getModifiers()) &&
+                    ((fields[i].getType() == Hashtable.class) || (fields[i].getType() == HashMap.class)))
+                {
+                    fields[i].setAccessible(true);
+                    if (fields[i].getType() == Hashtable.class)
+                    {
+                        Hashtable cache = (Hashtable) fields[i].get(null);
+                        if (cache != null)
+                        {
+                            cache.clear();
+                        }
+                    }
+                    else
+                    {
+                        HashMap cache = (HashMap) fields[i].get(null);
+                        if (cache != null)
+                        {
+                            cache.clear();
+                        }
+                    }
+                }
+            }
+        }
+    }
+
     private static class Actions implements PrivilegedExceptionAction
     {
-        public static final int GET_PROPERTY_ACTION = 0;
-        public static final int FOR_NAME_ACTION = 1;
-        public static final int CREATE_URL_ACTION = 2;
-        public static final int CREATE_URL_WITH_CONTEXT_ACTION = 3;
-        public static final int GET_ABSOLUTE_PATH_ACTION = 4;
-        public static final int FILE_EXISTS_ACTION = 5;
-        public static final int FILE_IS_DIRECTORY_ACTION = 6;
-        public static final int MAKE_DIRECTORY_ACTION = 7;
-        public static final int MAKE_DIRECTORIES_ACTION = 8;
-        public static final int LIST_DIRECTORY_ACTION = 9;
-        public static final int RENAME_FILE_ACTION = 10;
-        public static final int GET_FILE_INPUT_ACTION = 11;
-        public static final int GET_FILE_OUTPUT_ACTION = 12;
-        public static final int DELETE_FILE_ACTION = 13;
-        public static final int OPEN_JARX_ACTION = 14;
-        public static final int GET_URL_INPUT_ACTION = 15;
-        public static final int CREATE_MODULECLASSLOADER_ACTION = 16;
-        public static final int START_ACTIVATOR_ACTION = 17;
-        public static final int STOP_ACTIVATOR_ACTION = 18;
-        public static final int SYSTEM_EXIT_ACTION = 19;
-        public static final int OPEN_JARX_VERIFY_ACTION = 20;
-        public static final int GET_POLICY_ACTION = 21;
-        public static final int CREATE_TMPFILE_ACTION = 22;
-        public static final int OPEN_URLCONNECTION_ACTION = 23;
-        public static final int OPEN_JARURLCONNECTIONJAR_ACTION = 24;
-        public static final int ADD_EXTENSION_URL = 25;
-        public static final int GET_CONSTRUCTOR_ACTION = 26;
-        public static final int GET_METHOD_ACTION = 27;
-        public static final int INVOKE_METHOD_ACTION = 28;
-        public static final int INVOKE_CONSTRUCTOR_ACTION = 29;
-        public static final int SWAP_FIELD_ACTION = 30;
-        public static final int GET_FIELD_ACTION = 31;
-        public static final int GET_DECLAREDMETHOD_ACTION = 32;
+        public static final int INITIALIZE_CONTEXT = 0;
+        public static final int ADD_EXTENSION_URL = 1;
+        public static final int CREATE_TMPFILE_ACTION = 2;
+        public static final int CREATE_URL_ACTION = 3;
+        public static final int CREATE_URL_WITH_CONTEXT_ACTION = 4;
+        public static final int DELETE_FILE_ACTION = 5;
+        public static final int EXEC_ACTION = 6;
+        public static final int FILE_EXISTS_ACTION = 7;
+        public static final int FILE_IS_DIRECTORY_ACTION = 8;
+        public static final int FOR_NAME_ACTION = 9;
+        public static final int GET_ABSOLUTE_PATH_ACTION = 10;
+        public static final int GET_CONSTRUCTOR_ACTION = 11;
+        public static final int GET_DECLARED_CONSTRUCTOR_ACTION = 12;
+        public static final int GET_DECLARED_METHOD_ACTION = 13;
+        public static final int GET_FIELD_ACTION = 14;
+        public static final int GET_FILE_INPUT_ACTION = 15;
+        public static final int GET_FILE_OUTPUT_ACTION = 16;
+        public static final int GET_METHOD_ACTION = 17;
+        public static final int GET_POLICY_ACTION = 18;
+        public static final int GET_PROPERTY_ACTION = 19;
+        public static final int GET_PARENT_CLASS_LOADER_ACTION = 20;
+        public static final int GET_SYSTEM_CLASS_LOADER_ACTION = 21;
+        public static final int GET_URL_INPUT_ACTION = 22;
+        public static final int INVOKE_CONSTRUCTOR_ACTION = 23;
+        public static final int INVOKE_DIRECTMETHOD_ACTION = 24;
+        public static final int INVOKE_METHOD_ACTION = 25;
+        public static final int LIST_DIRECTORY_ACTION = 26;
+        public static final int MAKE_DIRECTORIES_ACTION = 27;
+        public static final int MAKE_DIRECTORY_ACTION = 28;
+        public static final int OPEN_JARX_ACTION = 29;
+        public static final int OPEN_JARX_VERIFY_ACTION = 30;
+        public static final int OPEN_URLCONNECTION_ACTION = 31;
+        public static final int RENAME_FILE_ACTION = 32;
         public static final int SET_ACCESSIBLE_ACTION = 33;
-        public static final int INVOKE_DIRECTMETHOD_ACTION = 34;
+        public static final int START_ACTIVATOR_ACTION = 34;
+        public static final int STOP_ACTIVATOR_ACTION = 35;
+        public static final int SWAP_FIELD_ACTION = 36;
+        public static final int SYSTEM_EXIT_ACTION = 37;
+        public static final int FLUSH_FIELD_ACTION = 38;
+        public static final int GET_CLASS_LOADER_ACTION = 39;
 
         private int m_action = -1;
         private Object m_arg1 = null;
@@ -970,6 +1094,11 @@ public class SecureAction
         private Object m_arg4 = null;
         private Object m_arg5 = null;
 
+        public void set(int action)
+        {
+            m_action = action;
+        }
+
         public void set(int action, Object arg1)
         {
             m_action = action;
@@ -1030,13 +1159,25 @@ public class SecureAction
             Object arg3 = m_arg3;
             Object arg4 = m_arg4;
             Object arg5 = m_arg5;
-            
+
             unset();
-            
-            if (action == GET_PROPERTY_ACTION)
+
+            if (action == INITIALIZE_CONTEXT)
+            {
+                return AccessController.getContext();
+            }
+            else if (action == GET_PROPERTY_ACTION)
             {
                 return System.getProperty((String) arg1, (String) arg2);
             }
+            else if (action == GET_PARENT_CLASS_LOADER_ACTION)
+            {
+                return ((ClassLoader) arg1).getParent();
+            }
+            else if (action == GET_SYSTEM_CLASS_LOADER_ACTION)
+            {
+                return ClassLoader.getSystemClassLoader();
+            }
             else if (action == FOR_NAME_ACTION)
             {
                 return Class.forName((String) arg1);
@@ -1052,6 +1193,10 @@ public class SecureAction
                 return new URL((URL) arg1, (String) arg2,
                     (URLStreamHandler) arg3);
             }
+            else if (action == EXEC_ACTION)
+            {
+                return Runtime.getRuntime().exec((String) arg1);
+            }
             else if (action == GET_ABSOLUTE_PATH_ACTION)
             {
                 return ((File) arg1).getAbsolutePath();
@@ -1104,11 +1249,6 @@ public class SecureAction
             {
                 return ((URLConnection) arg1).getInputStream();
             }
-// TODO: REFACTOR - SecureAction fix needed.
-//            else if (action == CREATE_MODULECLASSLOADER_ACTION)
-//            {
-//                return new ModuleClassLoader((ModuleImpl) arg1, (ProtectionDomain) arg2);
-//            }
             else if (action == START_ACTIVATOR_ACTION)
             {
                 ((BundleActivator) arg1).start((BundleContext) arg2);
@@ -1136,10 +1276,6 @@ public class SecureAction
             {
                 return ((URL) arg1).openConnection();
             }
-            else if (action == OPEN_JARURLCONNECTIONJAR_ACTION)
-            {
-                return ((JarURLConnection) arg1).getJarFile();
-            }
             else if (action == ADD_EXTENSION_URL)
             {
                 Method addURL =
@@ -1152,6 +1288,10 @@ public class SecureAction
             {
                 return ((Class) arg1).getConstructor((Class[]) arg2);
             }
+            else if (action == GET_DECLARED_CONSTRUCTOR_ACTION)
+            {
+                return ((Class) arg1).getDeclaredConstructor((Class[]) arg2);
+            }
             else if (action == GET_METHOD_ACTION)
             {
                 return ((Class) arg1).getMethod((String) arg2, (Class[]) arg3);
@@ -1180,13 +1320,21 @@ public class SecureAction
                 field.setAccessible(true);
                 return field.get(arg3);
             }
-            else if (action == GET_DECLAREDMETHOD_ACTION)
+            else if (action == GET_DECLARED_METHOD_ACTION)
             {
                 return ((Class) arg1).getDeclaredMethod((String) arg2, (Class[]) arg3);
             }
             else if (action == SET_ACCESSIBLE_ACTION)
             {
-                ((Method) arg1).setAccessible(true);
+                ((AccessibleObject) arg1).setAccessible(true);
+            }
+            else if (action == FLUSH_FIELD_ACTION)
+            {
+                _flush(((Class) arg1), arg2);
+            }
+            else if (action == GET_CLASS_LOADER_ACTION)
+            {
+                return ((Class) arg1).getClassLoader();
             }
 
             return null;
